//  Body.h - a rigid body.
//
//  Copyright (C) 2001--2002 Sam Varner
//
//  This file is part of Vamos Automotive Simulator.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef _BODY_H_
#define _BODY_H_

#include <vamos/geometry/Three_Vector.h>
#include <vamos/geometry/Three_Matrix.h>
#include <vamos/geometry/Surface.h>
#include <vamos/geometry/Inertia_Tensor.h>
#include <vamos/body/Frame.h>
#include <vamos/body/Particle.h>

#include <vector>

using Vamos_Geometry::Three_Vector;
using Vamos_Geometry::Three_Matrix;
using Vamos_Geometry::Inertia_Tensor;

namespace Vamos_Body
{
  //* A rigid body.
  class Body : public Frame
  {
	// The body's initial position vector, used by reset ().
	Three_Vector m_initial_position;

	Three_Vector m_last_position;

	// The velocity of the center of mass.
	Three_Vector m_cm_velocity;

	Three_Vector m_last_cm_velocity;

	Three_Vector m_last_velocity;

	Three_Matrix m_last_orientation;

	Three_Vector m_last_ang_velocity;

	// The acceleration due to gravity, distance/time^2.  The units
	// determine the distance and time units used in the rest of the
	// simulation.
	Three_Vector m_gravity;

	// The elapsed time since the last time step.
	double m_delta_time;

	// The total mass of the body.
	double m_mass;

	// Common code for the two reset () methods.
	void private_reset ();

  protected:
	// The inertia tensor for the body.
	Inertia_Tensor m_inertia;

	// A vector of pointers to the force and torque producers that
	// make up the body
	std::vector <Particle*> m_particles;

	// The position of the center of mass of the body relative to the
	// origin of the body.
	Three_Vector m_body_cm;

	// Return the inertia of the INDEXth particle for a force in the 
	// direction of the vector FORCE_DIRECTION.
	double inertia (int index, const Three_Vector& force_direction);

  public:
	//** Constructors

	// Specify the position and orientation of the body.
	Body (const Three_Vector& pos, const Three_Matrix& orient);

	// Specify the position, the orientation is the same as the parent.
	Body (const Three_Vector& pos);

	// The body is coincident with the parent.
	Body ();


	//** Destructor
	virtual ~Body ();

	// Return the position of the center of mass of the body with
	// respect to the world.
	Three_Vector cm_position ();

	// Return the position of the INDEXth particle of `m_particles'
	// with respect to the world.
	Three_Vector position (int index);

	// Provide access to the frame's position method.  Return the
	// origin of the body.
	Three_Vector position () const { return Frame::position (); }

	// Return the contact position of the INDEXth particle of
	// `m_particles' with respect to the world.
	Three_Vector contact_position (int index);

	// Return the smallest contact position z-value of the particles.
	double lowest_contact_position () const;

	// Add a particle to the body.
	void add_particle (Particle* const comp) 
	{ m_particles.push_back (comp); } 

	// Calculate the center of mass, the ineritia tensor, and its
	// inverse.
	void update_center_of_mass ();

	// Provide access to the particles.
	std::vector <Particle*>& particles () { return m_particles; }

	void find_forces ();

	// Advance the body in time by TIME.
	void propagate (double time);

	void rewind ();

	void end_timestep ();

	// Called by the world to tell the body what the acceleration due
	// to gravity is.
	void gravity (const Three_Vector& grav) { m_gravity = grav; }

	// Return the velocity of the INDEXth particle in the parent frame.
	// With no argument, return the velocity of the origin of the body
	// frame.
	Three_Vector velocity (int index = -1);

	// Return the velocity of the center of mass.
	Three_Vector cm_velocity () const { return m_cm_velocity; }

	// Set the velocity of the center of mass.
	void cm_velocity (Three_Vector vel) { m_cm_velocity = vel; }

	// Handle a collision.  The return value is how much the particle
	// has moved in the z-direction as a result of the contact.
	Three_Vector contact (int index, 
						  const Three_Vector& distance,
						  const Vamos_Geometry::Surface& surface);

	// Transform the wind into the body frame and send it to the 
	// INDEXth particle which must be an Aerodynamic_Device. 
	void wind (int index, const Three_Vector& wind_vector, double density);

	// Return the total mass.
	double mass () const { return m_mass; }

	// This function is defined by subclasses that work with a
	// graphics system.
	virtual void draw () {};

	// Return the body to its initial state at its initial position.
	virtual void reset ();

	// Return the body to its initial state at a particular position and
	// orientation.
	virtual void reset (const Three_Vector& position, 
						const Three_Matrix& orientation);
  };
}

#endif // !_BODY_H_
