//  Car.h - a body with wheels.
//
//  Copyright (C) 2001 Sam Varner
//
//  This file is part of Vamos Automotive Simulator.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef _CAR_H_
#define _CAR_H_

#include <vamos/geometry/Three_Vector.h>
#include <vamos/body/Body.h>
#include <vamos/body/Drivetrain.h>
#include <vamos/body/Fuel_Tank.h>
#include <vamos/body/Particle.h>
#include <vamos/body/Wheel.h>

#include <vector>

namespace Vamos_Body
{
  //* A class that handles gradual application of a control that's
  // operated by a button, such as the clutch.  If you're using a
  // keyboard instead of a joystick, it also handles steering, gas and
  // brake.
  class Key_Control
  {
	// The current setting of this control.
	double m_value;

	// The desired setting of this control.
	double m_target;

	// How fast `m_value' approaches `m_target'.
	double m_rate;

  public:
	//** Constructor
	Key_Control ();
	
	// Set the target setting of this control.  NEW_TARGET is the
	// desired setting.  TIME is how long it should take for the
	// setting to go from 0.0 to RANGE.  `m_rate' is calculated in
	// this function.
	void target (double new_target, double time = 0.0, double range = 1.0);

	// Update the setting of this control.  The setting move toward
	// `m_target' by the ammount `m_rate' * TIME.
	double update (double time);

	// Return the current value of this control.
	double value () const { return m_value; }
  };


  //* A body with wheels.
  class Car : public Body
  {
  protected:
	// A pointer to the car's drivetrain.
	Drivetrain* mp_drivetrain;

	// A pointer to the car's fuel tank.
	Fuel_Tank* mp_fuel_tank;

	// The fraction of braking torque applied to the front wheels.
	double m_front_brake_bias;

	// The maximum angle for the steered wheels.
	double m_max_steer_angle;

	// The sum of the sliding speeds of the tires.
	double m_slide;

	// True if a shift has been requested but not yet made due to the
	// delay `m_shift_delay'.
	bool m_shift_pending;

	// The amount of time elapsed after the shift request.
	double m_shift_timer;

	// How long to wait between getting a shift request and actually
	// shifting. 
	double m_shift_delay;

	// The gear to shift to when `m_shift_timer' reaches
	// `m_shift_delay'. 
	int m_new_gear;

	// The control that gradually applies steering when the keyboard
	// is used.
	Key_Control m_steer_key_control;

	// The control that gradually applies the throttle when the
	// keyboard is used.
	Key_Control m_gas_key_control;

	// The control that gradually applies braking when the keyboard
	// is used.
	Key_Control m_brake_key_control;

	// The control that gradually applies the clutch when the keyboard
	// is used.
	Key_Control m_clutch_key_control;

	// A pointer to the frontmost particle of the car.
	Particle* mp_front_particle;

  public:
	//** Constructor
	// The car is responsible for deleting the drivetrain.  All other
	// arguments are Particles that are deleted by the Body.
	Car (const Vamos_Geometry::Three_Vector& pos,
		 Drivetrain* drivetrain, 
		 Fuel_Tank* fuel_tank,
		 const std::vector <Particle*>& parts);

	//** Destructor
	virtual ~Car ();

	// Advance the car in time by TIME.
	virtual void propagate (double time);
  
	// Change the steering angle to ANGLE with a time constant of TIME.
	void steer (double angle, double time = 0.0);

	// Change the throttle to FACTOR with a time constant of TIME.
	void gas (double factor, double time = 0.0);

	// Change the brakes to FACTOR with a time constant of TIME.
	void brake (double factor, double time = 0.0);

	// Shift to the next lower gear after a delay of DELAY.  The gear
	// to be chosen is returned.
	int shift_down (double delay = 0.0);

	// Shift to the next higher gear after a delay of DELAY.  The gear
	// to be chosen is returned.
	int shift_up (double delay = 0.0);

	// Shift to GEAR after a delay of DELAY.  The gear to be chosen is
	// returned.
	int shift (int gear, double delay = 0.0);

	// Engage the clutch with a time constant of TIME.
	void engage_clutch (double time);

	// Disengage the clutch with a time constant of TIME.
	void disengage_clutch (double time);

	// Start the engine.
	void start () { mp_drivetrain->engine ().start (); }
  
	// Set the front brake bias to BIAS.
	void brake_bias (double bias);

	// Set the largest possible steering angle.
	void max_steer_angle (double degree_angle) 
	{ m_max_steer_angle = degree_angle; }
	
	// Return the sum of the sliding speeds of the tires.
	double slide () const { return m_slide; }

	// Return the pointer to the WHEEL_INDEXth wheel.
	Wheel* wheel (int wheel_index) const;

	// Return the pointer to the front-most particle.
	Particle* front_particle () const { return mp_front_particle; }

	// Return a reference to the engine.
	Engine& engine () { return mp_drivetrain->engine (); }

	// Return a reference to the transmission.
	Transmission& transmission () { return mp_drivetrain->transmission (); }

	// Return a reference to the fuel tank.
	Fuel_Tank& fuel_tank () { return *mp_fuel_tank; }

	// Return the last gear to be selected.  The shift may not have
	// occurred yet due to the shift delay spicfied in the call to
	// `shift_up ()', `shift_down ()', or `shift ()'.
	int gear () const { return m_new_gear; }

	// Handle soft collisions.  The return value is how much the
	// particle has moved in the z-direction as a result of the
	// contact.  This function is called by a World object.  INDEX -
	// the index for the particle that made contact in the particle
	// vector.  DISTANCE - the normal vector to the particle that made
	// contact ANG_VELOCITY - the component of the car's angular
	// velocity about DISTANCE.  SURFACE describes the surface where
	// contact was made.
	Three_Vector wheel_contact (int index, 
								const Three_Vector& distance,
								const Vamos_Geometry::Surface& surface);

	//** Enhance Body's reset methods by re-initializing the engine
	// and gearbox.

	// Restore the initial conditions.
	void reset ();

	// Restore the initial conditions and then set the position to
	// POSITION and the orientation to ORIENTATION.
	void reset (const Three_Vector& position, const Three_Matrix& orientation);
  };
}

#endif // !_CAR_H_
